# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrapper of the Token Classification models on HuggingFace platform (context
understanding)
"""
import re
from typing import Dict, Set, List, Optional, Tuple, Any
from transformers import pipeline

from forte.common import Resources
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor
from forte.common import ProcessorConfigError
from forte.utils.tagging_scheme import bio_merge
from forte.utils import get_class

__all__ = [
    "TokenClassification",
]


class TokenClassification(PackProcessor):
    r"""Wrapper of the models on HuggingFace platform with pipeline tag of
    `ner`.
    https://huggingface.co/models?pipeline_tag=token-classification
    This wrapper could take the models on HuggingFace platform with pipeline
    tag of `ner` to make prediction on token classification tasks, for example,
    NER, POS tagging, word segmentation, etc.
    This processor takes user specified entry type in the input pack and the
    prediction result goes to the user specified attribute name of that entry
    type in the data pack.

    For example, if users want to do NER task on each sentence, and get BIO
    tagging merged, and store results as the attribute `ner_type` in each
    EntityMention, then the following config could be used:
        "entry_type": "ft.onto.base_ontology.Sentence",
        "output_entry_type": "ft.onto.base_ontology.EntityMention",
        "attribute_name": "ner",
        "tagging_scheme": "bio-merge"

    If user want to do POS tagging on each sentence and store results
    as the attribute `pos` in each Token, then the following config could
    be used:
        "entry_type": "ft.onto.base_ontology.Sentence",
        "output_entry_type": "ft.onto.base_ontology.Token",
        "attribute_name": "pos",
        "tagging_scheme": "no-merge"


    if user want to do word segmentation task on each sentence, and do BIO
    tagging merged, and store the results as the attribute `word_segment` in
    each Token, then the following config could be used:
        "entry_type": "ft.onto.base_ontology.Sentence",
        "output_entry_type": "ft.onto.base_ontology.Token",
        "attribute_name": "word_segment",
        "tagging_scheme": "bio-merge"

    For other settings, please refer to the `default_config` function for more
    information.

    """

    def __init__(self):
        super().__init__()
        self.classifier = None

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.classifier = pipeline(
            "ner",
            model=self.configs.model_name,
            tokenizer=self.configs.tokenizer,
            framework=self.configs.framework,
            device=self.configs.cuda_device,
        )

    def _process(self, input_pack: DataPack):
        """Perform HuggingFace NER Pipeline on the input data pack.

        Args:
            input_pack: Input pack to fill
        Returns:
        """
        if not self.configs.entry_type:
            raise ProcessorConfigError("Please specify an input entry type!")

        output_entry = get_class(self.configs.output_entry_type)

        for entry_specified in input_pack.get(self.configs.entry_type):
            result = self.classifier(entry_specified.text)

            if self.configs.tagging_scheme == "bio-merge":  # Merge BIO tagging
                result_types, result_indices = self._merge_bio_tokens(result)

            elif self.configs.tagging_scheme == "no-merge":
                result_indices = []
                result_types = []
                for token in result:
                    start, end = token["start"], token["end"]
                    result_types.append(token["entity"])
                    result_indices.append((start, end))
            else:
                raise ProcessorConfigError(
                    f"The tagging_scheme strategy {self.configs.tagging_scheme}"
                    f"was not defined. Please check your input config."
                )

            for type, (start, end) in zip(result_types, result_indices):
                entity = output_entry(
                    pack=input_pack,
                    begin=entry_specified.span.begin + int(start),
                    end=entry_specified.span.begin + int(end),
                )
                setattr(entity, self.configs.attribute_name, type)

    def _merge_bio_tokens(
        self, tokens: List[Dict[str, Any]]
    ) -> Tuple[List[Optional[str]], List[Tuple[int, int]]]:
        """Perform token merge on BIO tagging format.

        Args:
            tokens: the output result from model
        Returns:
            result_types: list of merged entity type.
            result_indices: list of (start, end) index for the merged entities.

        """
        indices = [(token["start"], token["end"]) for token in tokens]

        tags: List[str] = []
        types: List[str] = []

        for token in tokens:
            if re.match("^[B|I]-.*", token["entity"]):
                tag, type = token["entity"].split("-")
            elif token["entity"] in ["B", "I"]:
                tag, type = token["entity"], ""
            else:
                raise ValueError("cannot process the BIO pattern.")
            tags.append(tag)
            types.append(type)

        result_types, result_indices = bio_merge(tags, types, indices)
        return result_types, result_indices

    @classmethod
    def default_configs(cls):
        r"""This defines a basic config structure for TokenClassification.

        Following are the keys for this dictionary:
            - `entry_type`: defines which entry type in the input pack to make
              prediction on. The default makes prediction on each `Sentence`
              in the input pack.
            - `output_entry_type`: defines which entry type in the input pack
              that the prediction should be saved as output. The default
              saves prediction on `Token` in the input pack.
            - `attribute_name`: defines which attribute of the
              `output_entry_type` in the input pack to save prediction to.
              The default saves prediction to the `ner` attribute for each
              `output_entry_type` in the input pack.
            - `tagging_scheme`: defines the result merged tagging_scheme,
              for example, "no-merge" to output the original result,
              "bio-merge" to merge the BIO tagging format result.
              The default is "no-merge".
            - `model_name`: language model, default is `"dslim/bert-base-NER"`.
              The wrapper supports HuggingFace models with pipeline tag of
              `ner`.
            - `tokenizer`: language model for tokenization, default is
              `"dslim/bert-base-NER"`.
            - `framework`: The framework of the model, should be `"pt"` or
              `"tf"`. This information should be obtained from Huggingface
              model hub.
            - `cuda_device`: Device ordinal for CPU/GPU supports. Setting
              this to -1 will leverage CPU, a positive will run the model
              on the associated CUDA device id. Default is -1.

        Returns: A dictionary with the default config for this processor.
        """
        return {
            "entry_type": "ft.onto.base_ontology.Sentence",
            "output_entry_type": "ft.onto.base_ontology.EntityMention",
            "attribute_name": "ner",
            "tagging_scheme": "bio-merge",  # "no-merge"
            "model_name": "dslim/bert-base-NER",
            "tokenizer": "dslim/bert-base-NER",
            "framework": "pt",
            "cuda_device": -1,
        }

    def expected_types_and_attributes(self):
        r"""Method to add expected type `ft.onto.base_ontology.Sentence` which
        would be checked before running the processor if
        the pipeline is initialized with
        `enforce_consistency=True` or
        :meth:`~forte.pipeline.Pipeline.enforce_consistency` was enabled for
        the pipeline.
        """
        return {self.configs["entry_type"]: set()}

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `output_entry_type`, which is
        user specified entry type with user specified attribute name
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """

        if self.configs.output_entry_type not in record_meta.keys():
            record_meta[self.configs.output_entry_type] = set()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Any, Set

from nltk.tokenize.util import align_tokens
from transformers import AutoTokenizer
from ft.onto.base_ontology import Subword

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor

__all__ = [
    "BERTTokenizer",
]


class BERTTokenizer(PackProcessor):
    r"""A wrapper of BERT tokenizer."""

    def __init__(self):
        super().__init__()
        self.tokenizer = None

    # pylint: disable=unused-argument
    def initialize(self, resources: Resources, configs: Config):
        self.tokenizer = AutoTokenizer.from_pretrained(configs.model_path)

    def _process(self, input_pack: DataPack):
        inputs = self.tokenizer(input_pack.text, return_tensors="pt")
        tokens = self.tokenizer.convert_ids_to_tokens(
            inputs["input_ids"][0].tolist()
        )[1:-1]
        tokens_clean = [
            token.replace("##", "") if token.startswith("##") else token
            for token in tokens
        ]

        for i, (begin, end) in enumerate(
            align_tokens(tokens_clean, input_pack.text.lower())
        ):
            subword = Subword(input_pack, begin, end)
            subword.is_first_segment = not tokens[i].startswith("##")

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        r"""Returns a `dict` of configurations of the processor with default
        values. Used to replace the missing values of input ``configs`` during
        pipeline construction.
        """
        return {"model_path": None}

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type `ft.onto.base_ontology.Subword`
        of current processor `BERTTokenizer`
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        record_meta["ft.onto.base_ontology.Subword"] = {"is_first_segment"}



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrapper of the Question Answering models on HuggingFace platform (context
understanding)
"""
from transformers import pipeline
from ft.onto.base_ontology import Phrase, Sentence

from forte.common import Resources
from forte.common.configuration import Config
from forte.data.multi_pack import MultiPack, MultiPackLink
from forte.processors.base import MultiPackProcessor

__all__ = [
    "QuestionAnsweringMulti",
]


class QuestionAnsweringMulti(MultiPackProcessor):
    r"""Wrapper of the models on HuggingFace platform with pipeline tag of
    `question-answering` (reading comprehension).
    https://huggingface.co/models?pipeline_tag=question-answering
    This wrapper could take any model name on HuggingFace platform with pipeline
    tag of `question-answering` in configs to make prediction on the context of
    user specified entry type in the multiple input pack and the prediction
    result would be annotated as `Phrase` in the output pack, which would be
    linked to the question pack by `MultiPackLink`.
    """

    def __init__(self):
        super().__init__()
        self.extractor = None

    def set_up(self):
        device_num = self.configs["cuda_devices"]
        self.extractor = pipeline(
            "question-answering",
            model=self.configs.model_name,
            tokenizer=self.configs.model_name,
            framework="pt",
            device=device_num,
        )

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.set_up()

    def _process(self, input_pack: MultiPack):
        context_list = []
        doc_id_list = []
        for doc_id in input_pack.pack_names:
            if doc_id == self.configs.question_pack_name:
                continue

            pack = input_pack.get_pack(doc_id)
            context_list.append(pack.get_single(self.configs.entry_type).text)
            doc_id_list.append(doc_id)

        question_pack = input_pack.get_pack(self.configs.question_pack_name)
        first_question = question_pack.get_single(Sentence)
        question_list = [question_pack.text for i in range(len(context_list))]
        result_collection = self.extractor(
            context=context_list,
            question=question_list,
            max_answer_len=self.configs.max_answer_len,
            handle_impossible_answer=self.configs.handle_impossible_answer,
        )
        for i, result in enumerate(result_collection):
            start = result["start"]
            end = result["end"]
            doc_pack = input_pack.get_pack(doc_id_list[i])
            ans_phrase = Phrase(pack=doc_pack, begin=start, end=end)
            input_pack.add_entry(
                MultiPackLink(input_pack, first_question, ans_phrase)
            )

    @classmethod
    def default_configs(cls):
        r"""This defines a basic config structure for `QuestionAnsweringSingle`.

        Following are the keys for this dictionary:
            - `entry_type`: defines which entry type in the input pack to make
              prediction on. The default makes prediction on each `Document`
              in the input pack.
            - `model_name`: language model, default is
              `"ktrapeznikov/biobert_v1.1_pubmed_squad_v2"`.
              The wrapper supports Hugging Face models with pipeline tag of
              `question-answering`.
            - `max_answer_len`: The maximum length of predicted answers (e.g.,
              only answers with a shorter length are considered).
            - `cuda_device`: Device ordinal for CPU/GPU supports. Setting
              this to -1 will leverage CPU, a positive will run the model
              on the associated CUDA device id.
            - `handle_impossible_answer`: Whether or not we accept
              impossible as an answer.

        Returns: A dictionary with the default config for this processor.
        """
        return {
            "question_pack_name": "question",
            "entry_type": "ft.onto.base_ontology.Document",
            "model_name": "ktrapeznikov/biobert_v1.1_pubmed_squad_v2",
            "max_answer_len": 15,
            "cuda_devices": -1,
            "handle_impossible_answer": False,
        }



# pylint: disable=logging-fstring-interpolation
from typing import Dict, List, Optional, Tuple, Any, Set
import logging
import numpy as np
import torch

from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.ontology.top import Annotation
from forte.processors.base.batch_processor import RequestPackingProcessor
from ft.onto.base_ontology import EntityMention, Subword
from transformers import (
    AutoConfig,
    AutoModelForTokenClassification,
    AutoTokenizer,
)

__all__ = [
    "BioBERTNERPredictor",
]


class BioBERTNERPredictor(RequestPackingProcessor):
    """
    An Named Entity Recognizer fine-tuned on BioBERT

    Note that to use :class:`BioBERTNERPredictor`, the :attr:`ontology` of
    :class:`Pipeline` must be an ontology that include
    ``ft.onto.base_ontology.Subword`` and ``ft.onto.base_ontology.Sentence``.
    """

    def __init__(self):
        super().__init__()
        self.resources = None
        self.device = None

        self.ft_configs = None
        self.model_config = None
        self.model = None
        self.tokenizer = None

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if resources.get("device"):
            self.device = resources.get("device")
        else:
            self.device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )

        self.resources = resources
        self.ft_configs = configs

        model_path = self.ft_configs.model_path
        self.model_config = AutoConfig.from_pretrained(model_path)
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.model = AutoModelForTokenClassification.from_pretrained(
            model_path,
            from_tf=bool(".ckpt" in model_path),
            config=self.model_config,
        )
        self.model.to(self.device)

    @torch.no_grad()
    def predict(
        self, data_batch: Dict[str, Dict[str, List[str]]]
    ) -> Dict[str, Dict[str, List]]:  # numpy typing is removed due to
        # incompatibility issues due to upgrading from numpy verions 1.19 to
        # 1.21
        sentences = data_batch["context"]
        subwords = data_batch["Subword"]

        inputs = self.tokenizer(sentences, return_tensors="pt", padding=True)
        inputs = {key: value.to(self.device) for key, value in inputs.items()}

        input_shape = inputs["input_ids"].shape
        if input_shape[1] > 512:
            # TODO: Temporarily work around the length problem.
            #   The real solution should further split up the sentences to make
            #   the sentences shorter.
            labels_idx = (
                inputs["input_ids"]
                .new_full(input_shape, 2, device="cpu")[:, 1:-1]
                .numpy()
            )
        else:
            outputs = self.model(**inputs)[0].cpu().numpy()
            score = np.exp(outputs) / np.exp(outputs).sum(-1, keepdims=True)
            labels_idx = score.argmax(axis=-1)[:, 1:-1]  # Remove placeholders.

        pred: Dict = {"Subword": {"ner": [], "tid": []}}

        for i in range(len(subwords["tid"])):
            tids = subwords["tid"][i]
            ner_tags = []
            for j in range(len(tids)):
                ner_tags.append(self.model.config.id2label[labels_idx[i, j]])

            pred["Subword"]["ner"].append(np.array(ner_tags))
            pred["Subword"]["tid"].append(np.array(tids))

        return pred

    def _complete_entity(
        self,
        subword_entities: List[Dict[str, Any]],
        data_pack: DataPack,
        tids: List[int],
    ) -> Tuple[int, int]:
        """Complete entity span from predicted subword entities

        Start from the first subword with predicted entity. If this entity
        is a subword (e.g. "##on"), then move on to the previous subword until
        it's no longer a subword (e.g. "br")

        """

        first_idx: int = subword_entities[0]["idx"]
        first_tid = subword_entities[0]["tid"]
        while (
            first_idx > 0
            and not data_pack.get_entry(first_tid).is_first_segment
        ):
            first_idx -= 1
            first_tid = tids[first_idx]

        last_idx: int = subword_entities[-1]["idx"]
        while (
            last_idx < len(tids) - 1
            and not data_pack.get_entry(tids[last_idx + 1]).is_first_segment
        ):
            last_idx += 1

        return first_idx, last_idx

    def _compose_entities(
        self,
        entities: List[Dict[str, Any]],
        data_pack: DataPack,
        tids: List[int],
    ) -> List[Tuple[int, int]]:
        """Composes entity spans from subword entity predictions

        Label Syntax:
        A "B" label indicates the beginning of an entity, an "I" label
        indicates the continuation of an entity, and an "O" label indicates
        the absence of an entity.
        Example: with - br - ##on - ##chi - ##oli - ##tis - .
                 O    - B  - I    - I     - I     - I     - O

        Due to possible instabilities of the model on out-of-distribution data,
        sometimes the prediction may not follow the label format.
        Example 1: with - br - ##on - ##chi - ##oli - ##tis - .
                   O    - B  - I    - O     - I     - O     - O

        Example 2: with - br - ##on - ##chi - ##oli - ##tis - .
                   O    - O  - O    - I     - I     - I     - O

        This method takes entity predictions of subwords and recovers the
        set of complete entities, defined by the indices of their beginning
        and ending subwords. (begin_idx, end_idx)
        """

        complete_entities: List[Tuple[int, int]] = []
        subword_entities: List[Dict[str, Any]] = []

        for entity in entities:
            subword = data_pack.get_entry(entity["tid"])

            if entity["label"] == "B" and subword.is_first_segment:
                # Flush the existing entity and start a new entity
                if subword_entities:
                    complete_entity = self._complete_entity(
                        subword_entities, data_pack, tids
                    )
                    complete_entities.append(complete_entity)
                subword_entities = [entity]
            else:
                # Continue accumuulating subword entities
                subword_entities.append(entity)

        if subword_entities:
            complete_entity = self._complete_entity(
                subword_entities, data_pack, tids
            )
            complete_entities.append(complete_entity)

        return complete_entities

    def pack(
        self,
        data_pack: DataPack,
        output_dict: Optional[Dict[str, Dict[str, List[Any]]]] = None,
        context: Optional[Annotation] = None,
    ):
        """
        Write the prediction results back to datapack by aggregating subwords
        into named entity mentions.
        """
        if output_dict is None:
            return
        if context is not None:
            logging.warning("context parameter is not used in pack() method.")

        for i in range(len(output_dict["Subword"]["tid"])):
            tids = output_dict["Subword"]["tid"][i]
            labels = output_dict["Subword"]["ner"][i]

            # Filter to labels not in `self.ft_configs.ignore_labels`
            entities = [
                dict(idx=idx, label=label, tid=tid)
                for idx, (label, tid) in enumerate(zip(labels, tids))
                if label not in self.ft_configs.ignore_labels
            ]
            entity_groups = self._compose_entities(entities, data_pack, tids)
            # Add NER tags and create EntityMention ontologies.
            for first_idx, last_idx in entity_groups:
                first_token: Subword = data_pack.get_entry(tids[first_idx])
                begin = first_token.span.begin

                last_token: Subword = data_pack.get_entry(tids[last_idx])
                end = last_token.span.end
                entity = EntityMention(data_pack, begin, end)
                entity.ner_type = self.ft_configs.ner_type

    @classmethod
    def default_configs(cls):
        r"""Default config for NER Predictor"""
        return {
            "model_path": None,
            "ner_type": "BioEntity",
            "ignore_labels": ["O"],
            "batcher": {
                "batch_size": 10,
                "context_type": "ft.onto.base_ontology.Sentence",
                "requests": {
                    "ft.onto.base_ontology.Subword": [],
                    "ft.onto.base_ontology.Sentence": [],
                },
            },
        }

    # @staticmethod
    # def _define_context() -> Type[Annotation]:
    #     return Sentence
    #
    # @staticmethod
    # def _define_input_info() -> DataRequest:
    #     input_info: DataRequest = {
    #         Subword: [],
    #         Sentence: [],
    #     }
    #     return input_info

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of current processor
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        record_meta["ft.onto.base_ontology.EntityMention"] = {"ner_type"}

    def expected_types_and_attributes(self):
        r"""Method to add expected type ft.onto.base_ontology.Subword` with
        attribute `is_first_segment` and `ft.onto.base_ontology.Sentence` which
        would be checked before running the processor if
        the pipeline is initialized with
        `enforce_consistency=True` or
        :meth:`~forte.pipeline.Pipeline.enforce_consistency` was enabled for
        the pipeline.
        """
        return {
            "ft.onto.base_ontology.Subword": {"is_first_segment"},
            "ft.onto.base_ontology.Sentence": set(),
        }



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Wrapper of the Zero Shot Classifier models on HuggingFace platform
"""
from typing import Dict, Set
import importlib

from transformers import pipeline

from forte.common import Resources
from forte.common.configuration import Config
from forte.data.data_pack import DataPack
from forte.processors.base import PackProcessor

__all__ = [
    "ZeroShotClassifier",
]


class ZeroShotClassifier(PackProcessor):
    r"""Wrapper of the models on HuggingFace platform with pipeline tag of
    `zero-shot-classification`.
    https://huggingface.co/models?pipeline_tag=zero-shot-classification
    This wrapper could take any model name on HuggingFace platform with pipeline
    tag of `zero-shot-classification` in configs to make prediction on the user
    specified entry type in the input pack and the prediction result goes to the
    user specified attribute name of that entry type in the output pack. User
    could input the prediction labels in the config with any word or phrase.

    """

    def __init__(self):
        super().__init__()
        self.classifier = None

    def set_up(self):
        device_num = self.configs["cuda_device"]
        self.classifier = pipeline(
            "zero-shot-classification",
            model=self.configs.model_name,
            framework="pt",
            device=device_num,
        )

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)
        self.set_up()

    def _process(self, input_pack: DataPack):
        path_str, module_str = self.configs.entry_type.rsplit(".", 1)

        mod = importlib.import_module(path_str)
        entry = getattr(mod, module_str)
        for entry_specified in input_pack.get(entry_type=entry):
            result = self.classifier(
                sequences=entry_specified.text,
                candidate_labels=self.configs.candidate_labels,
                hypothesis_template=self.configs.hypothesis_template,
                multi_class=self.configs.multi_class,
            )
            curr_dict = getattr(entry_specified, self.configs.attribute_name)
            for idx, lab in enumerate(result["labels"]):
                curr_dict[lab] = round(result["scores"][idx], 4)
            setattr(entry_specified, self.configs.attribute_name, curr_dict)

    @classmethod
    def default_configs(cls):
        r"""This defines a basic config structure for ZeroShotClassifier.

        Following are the keys for this dictionary:
            - `entry_type`: defines which entry type in the input pack to make
              prediction on. The default makes prediction on each `Sentence`
              in the input pack.
            - `attribute_name`: defines which attribute of the `entry_type`
              in the input pack to save prediction to. The default
              saves prediction to the `classification` attribute for each
              `Sentence` in the input pack.
            - `multi_class`: whether to allow multiple class true
            - `model_name`: language model, default is
              `"valhalla/distilbart-mnli-12-1"`.
              The wrapper supports Hugging Face models with pipeline tag of
              `zero-shot-classification`.
            - `candidate_labels`: The set of possible class labels to
              classify each sequence into. Can be a single label, a string of
              comma-separated labels, or a list of labels. Note that for the
              model with a specific language, the candidate_labels need to
              be of that language.
            - `hypothesis_template`: The template used to turn each label
              into an NLI-style hypothesis. This template must include a {}
              or similar syntax for the candidate label to be inserted into
              the template. For example, the default
              template is :obj:`"This example is {}."` Note that for the
              model with a specific language, the hypothesis_template need to
              be of that language.
            - `cuda_device`: Device ordinal for CPU/GPU supports. Setting
              this to -1 will leverage CPU, a positive will run the model
              on the associated CUDA device id.

        Returns: A dictionary with the default config for this processor.
        """
        return {
            "entry_type": "ft.onto.base_ontology.Sentence",
            "attribute_name": "classification",
            "multi_class": True,
            "model_name": "valhalla/distilbart-mnli-12-1",
            "candidate_labels": [
                "travel",
                "cooking",
                "dancing",
                "exploration",
            ],
            "hypothesis_template": "This example is {}.",
            "cuda_device": -1,
        }

    def expected_types_and_attributes(self):
        r"""Method to add expected type `ft.onto.base_ontology.Sentence` which
        would be checked before running the processor if
        the pipeline is initialized with
        `enforce_consistency=True` or
        :meth:`~forte.pipeline.Pipeline.enforce_consistency` was enabled for
        the pipeline.
        """
        return {self.configs["entry_type"]: set()}

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `ZeroShotClassifier` which is
        user specified entry type with user specified attribute name
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        if self.configs.entry_type in record_meta:
            record_meta[self.configs.entry_type].add(
                self.configs.attribute_name
            )
        else:
            record_meta[self.configs.entry_type] = {self.configs.attribute_name}



from fortex.huggingface.bio_ner_predictor import *
from fortex.huggingface.transformers_processor import *
from fortex.huggingface.question_and_answering_single import *
from fortex.huggingface.zero_shot_classifier import *
from fortex.huggingface.token_classification import *
